# -*- coding: utf-8 -*-
"""
Code for loading dataset.

It implements the function of returning train and test data of the following datasets
* the Boolean synthetic dataset
* Various UCI datasets through `get_tabular_dataset`
* MNIST, KMNIST, FashionMNIST etc.


There are also some dataset loading that is implemented but whose results are not presented currently
* The 3dshapes_xor datasets
* The CUB_200_2011 Bird identification dataset. adopted from https://github.com/TDeVries/cub2011_dataset and https://github.com/Muhammad-MujtabaSaeed/Caltech-Birds-Classification
The fecth-preprocess-loading function should be easy to understand and use.

"""


import glob
import numpy as np
from skimage.io import imread
import torch
from torch.utils.data import Dataset, DataLoader, TensorDataset
from torchvision import datasets, transforms
import os
import pickle
import pandas as pd
from torchvision.datasets.folder import default_loader
from torchvision.datasets.utils import download_url
from PIL import Image


def get_boolean_DNF_data(use_negation=False,random_seed=0,input_dim=10):
    np.random.seed(random_seed)
    n_samples=5000
    n_dim= input_dim
    n_DNF= 5
    n_condition=3
    if use_negation:
        groud_truth_DNF=np.zeros((n_DNF,n_dim*2))
    else:
        groud_truth_DNF=np.zeros((n_DNF,n_dim))

    for index, row in enumerate(groud_truth_DNF):
        temp = np.random.choice(n_dim, n_condition, replace=False)
        if use_negation:
            negation_or_not = np.random.choice([0,1], n_condition)
            for idx, i in enumerate(negation_or_not):
                if i == 1:
                    temp[idx]+=n_dim
        groud_truth_DNF[index,temp]=1
    # X = np.random.choice([0, 1], size=(n_samples,n_dim), p=[0.75, 0.25])
    X = np.random.choice([0, 1], size=(n_samples,n_dim), p=[0.5, 0.5])
    if use_negation:
        X_ = np.concatenate( (X,(1-X)),axis=1 )
    else:
        X_ = X
    conj = np.prod((1 - (1 - X_)[:, :, None] * groud_truth_DNF.transpose()[None, :, :]), axis=1)
    # self.conj = torch.mm((x), W.t())
    disj  = 1 - np.prod(1 - conj, axis=1)
    # print("temp",W)
    y = disj.reshape(-1,1)

    from sklearn.model_selection import train_test_split
    X_train, X_test, y_train, y_test = train_test_split(X, y, train_size=0.8)
    return X_train,y_train,X_test,y_test,groud_truth_DNF
    # return X,y,groud_truth_DNF

def get_tabular_dataset(dataset_name):
    from sklearn.model_selection import train_test_split
    from neural_dnf.utils import DBEncoder,read_csv
    import pandas as pd
    import os


    DATA_DIR = './dataset_storage/tabular_dataset'
    dataset=dataset_name
    if os.path.exists(DATA_DIR+"/tabular_dataset.pickle"):
        '''The discretization takes time. So if a cached file exist. Use the cached file'''
        print("using cached, already discretized UCI dataset")
        X_train_dict,y_train_dict,X_test_dict,y_test_dict = pickle.load( open( "./dataset_storage/tabular_dataset/tabular_dataset.pickle", "rb" ) )
        X_train = X_train_dict[dataset_name]
        y_train = y_train_dict[dataset_name]
        X_test = X_test_dict[dataset_name]
        y_test = y_test_dict[dataset_name]
    else:
        print('discretize now!')
        data_path = os.path.join(DATA_DIR, dataset + '.data')
        info_path = os.path.join(DATA_DIR, dataset + '.info')
        X_df, y_df, f_df, label_pos = read_csv(data_path, info_path, shuffle=True)

        X_train, X_test, y_train, y_test = train_test_split(X_df, y_df, train_size=0.8)

        db_enc = DBEncoder(f_df, discrete=True)
        # db_enc.fit(X_train[:1000], y_train[:1000])
        db_enc.fit(X_train, y_train)

        X_train, y_train = db_enc.transform(X_train, y_train)
        X_test, y_test = db_enc.transform(X_test, y_test)

        # by default the second digit of Y is the label we want
        N_feature=X_train.shape[-1]
        X_train=X_train[:,:N_feature]
        y_train=y_train[:,1].reshape(-1,1)
        X_test=X_test[:,:N_feature]
        y_test=y_test[:,1].reshape(-1,1)

    return X_train,y_train,X_test,y_test

def get_dataset_by_name(dataset_name,path_to_data_storage):
    from torch.utils.data import Dataset, DataLoader
    from torchvision import datasets, transforms
    path_to_data = path_to_data_storage + dataset_name
    if dataset_name == "MNIST":
        all_transforms = transforms.Compose([
            transforms.Resize(32),
            transforms.ToTensor()
        ])
        train_data = datasets.MNIST(path_to_data, train=True, download=True,
                                          transform=all_transforms)
        test_data = datasets.MNIST(path_to_data, train=False,
                                        transform=all_transforms)
        img_size=(1,32,32)
    elif dataset_name == "FashionMNIST":
        all_transforms = transforms.Compose([
            transforms.Resize(32),
            transforms.ToTensor()
        ])
        train_data = datasets.FashionMNIST(path_to_data, train=True, download=True,
                                          transform=all_transforms)
        test_data = datasets.FashionMNIST(path_to_data, train=False,
                                        transform=all_transforms)
        img_size=(1,32,32)
    elif dataset_name == "KMNIST":
        all_transforms = transforms.Compose([
            transforms.Resize(32),
            transforms.ToTensor()
        ])
        train_data = datasets.KMNIST(path_to_data, train=True, download=True,
                                          transform=all_transforms)
        test_data = datasets.KMNIST(path_to_data, train=False,
                                        transform=all_transforms)
        img_size=(1,32,32)
    elif dataset_name == "SVHN":
        all_transforms = transforms.Compose([
        transforms.Resize(64),
        transforms.ToTensor()
        ])
        train_data = datasets.SVHN(path_to_data, split='train', download=True,
                                          transform=all_transforms)
        test_data = datasets.SVHN(path_to_data, split='test', download=True,
                                        transform=all_transforms)
        img_size=(3,64,64)
    elif dataset_name == "CIFAR10":
        all_transforms = transforms.Compose([
        transforms.Resize(64),
        transforms.ToTensor()
        ])
        train_data = datasets.CIFAR10(path_to_data, train=True, download=True,
                                          transform=all_transforms)
        test_data = datasets.CIFAR10(path_to_data, train=False,
                                        transform=all_transforms)
        img_size=(3,64,64)
    else:
        '''UCI datasets'''
        print("Using UCI datasets",dataset_name)
        from dataloaders import get_tabular_dataset
        X_train,y_train,X_test,y_test = get_tabular_dataset(dataset_name)
        X_train = X_train.astype(np.float32)
        y_train = y_train.astype(np.float32)
        X = torch.tensor(X_train)
        y=torch.tensor(y_train)
        train_data=TensorDataset(X, y)
        X_test = X_test.astype(np.float32)
        y_test = y_test.astype(np.float32)
        X = torch.tensor(X_test)
        y=torch.tensor(y_test)
        test_data=TensorDataset(X, y)
        img_size = None






    return img_size,train_data,test_data

def get_mnist_dataloaders(batch_size=128, path_to_data='../data'):
    """MNIST dataloader with (1, 32, 32) images."""
    all_transforms = transforms.Compose([
        transforms.Resize(32),
        transforms.ToTensor(),
        # transforms.Normalize((0.1307,), (0.3081,))
    ])
    # all_transforms = transforms.Compose([
                           # transforms.ToTensor(),
                           # transforms.Normalize((0.1307,), (0.3081,))
                       # ])



    train_data = datasets.MNIST(path_to_data, train=True, download=True,
                                transform=all_transforms)
    test_data = datasets.MNIST(path_to_data, train=False,
                               transform=all_transforms)
    train_loader = DataLoader(train_data, batch_size=batch_size, shuffle=True)
    test_loader = DataLoader(test_data, batch_size=batch_size, shuffle=False)
    return train_loader, test_loader


def get_fashion_mnist_dataloaders(batch_size=128,
                                  path_to_data='../fashion_data'):
    """FashionMNIST dataloader with (32, 32) images."""
    all_transforms = transforms.Compose([
        transforms.Resize(32),
        transforms.ToTensor()
    ])
    train_data = datasets.FashionMNIST(path_to_data, train=True, download=True,
                                       transform=all_transforms)
    test_data = datasets.FashionMNIST(path_to_data, train=False,
                                      transform=all_transforms)
    train_loader = DataLoader(train_data, batch_size=batch_size, shuffle=True)
    test_loader = DataLoader(test_data, batch_size=batch_size, shuffle=False)
    return train_loader, test_loader

DATASET_BASE_PATH = "./dataset_storage/CUB_dataset"
PICKLE_PATH_TRAIN = "./dataset_storage/CUB_dataset/class_attr_data_10/train.pkl"
PICKLE_PATH_TEST = "./dataset_storage/CUB_dataset/class_attr_data_10/test.pkl"
# We are not using validation set

class CUBDataset(Dataset):
    """
    Returns a compatible Torch Dataset object customized for the CUB dataset
    """

    def __init__(self,   transform=None, train=True, use_attr=True, no_img=False, uncertain_label=False, n_class_attr=2, n_of_class=200):
        """
        Arguments:
        pkl_file_paths: list of full path to all the pkl data
        use_attr: whether to load the attributes (e.g. False for simple finetune)
        no_img: whether to load the images (e.g. False for A -> Y model)
        uncertain_label: if True, use 'uncertain_attribute_label' field (i.e. label weighted by uncertainty score, e.g. 1 & 3(probably) -> 0.75)
        image_dir: default = 'images'. Will be append to the parent dir
        n_class_attr: number of classes to predict for each attribute. If 3, then make a separate class for not visible
        transform: whether to apply any special transformation. Default = None, i.e. use standard ImageNet preprocessing
        """

        image_dir = "images"

        self.data = []
        # self.is_train = any(["train" in path for path in pkl_file_paths])
        if train is True:
            self.is_train = True
            pkl_file_paths = [PICKLE_PATH_TRAIN]
        else:
            self.is_train = False
            pkl_file_paths = [PICKLE_PATH_TEST]
        if not self.is_train:
            assert any([("test" in path) or ("val" in path) for path in pkl_file_paths])

        for file_path in pkl_file_paths:
            self.data.extend(pickle.load(open(file_path, 'rb')))

        self.transform = transform
        self.use_attr = use_attr
        self.no_img = no_img
        self.uncertain_label = uncertain_label
        self.image_dir = image_dir
        self.n_class_attr = n_class_attr
        self.n_of_class=n_of_class

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        img_data = self.data[idx]
        img_path = img_data['img_path']
        # Trim unnecessary paths
        try:
            idx = img_path.split('/').index('CUB_200_2011')
            if self.image_dir != 'images':
                img_path = '/'.join([self.image_dir] + img_path.split('/')[idx+1:])
                img_path = img_path.replace('images/', '')
            else:
                img_path = '/'.join(img_path.split('/')[idx:])
            img_path = DATASET_BASE_PATH + "/" + img_path
            img = Image.open(img_path).convert('RGB')
        except:
            img_path_split = img_path.split('/')
            split = 'train' if self.is_train else 'test'
            img_path = '/'.join(img_path_split[:2] + [split] + img_path_split[2:])
            img = Image.open(img_path).convert('RGB')

        class_label = img_data['class_label']
        if self.n_of_class != 200:
            print("error, not implemented, n_of_class must be 200")

        if self.transform:
            img = self.transform(img)

        if self.use_attr:
            if self.uncertain_label:
                attr_label = img_data['uncertain_attribute_label']
            else:
                attr_label = img_data['attribute_label']
            if self.no_img:
                if self.n_class_attr == 3:
                    N_ATTRIBUTES = 112
                    one_hot_attr_label = np.zeros((N_ATTRIBUTES, self.n_class_attr))
                    one_hot_attr_label[np.arange(N_ATTRIBUTES), attr_label] = 1
                    return one_hot_attr_label, class_label
                else:
                    attr_label = np.stack(attr_label).transpose()
                    return attr_label, class_label
            else:
                return img, class_label, attr_label
        else:
            return img, class_label

def get_bird_dataloader(resol=299,use_attr=True,no_img=False,n_of_class=200,batch_size = 64):
    '''
    resol: 299 the default value is for Inception.
    '''

    train_transform = transforms.Compose([
            transforms.ColorJitter(brightness=32/255, saturation=(0.5, 1.5)),
            transforms.RandomResizedCrop(resol),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(), #implicitly divides by 255
            transforms.Normalize(mean = [0.5, 0.5, 0.5], std = [2, 2, 2])
            ])
    test_transform = transforms.Compose([
            transforms.CenterCrop(resol),
            transforms.ToTensor(), #implicitly divides by 255
            transforms.Normalize(mean = [0.5, 0.5, 0.5], std = [2, 2, 2])
            ])

    train_dataset = CUBDataset(train=True,transform=train_transform,use_attr=use_attr,no_img=no_img,n_of_class=n_of_class)
    test_dataset = CUBDataset(train=False,transform=test_transform,use_attr=use_attr,no_img=no_img,n_of_class=n_of_class)


    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, drop_last=True, num_workers=4, pin_memory=False)
    test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=False, drop_last=False, num_workers=4, pin_memory=False)

    return train_loader,test_loader



def get_3dshapes_xor_dataset(n_samples_for_positive_class=6000,sample_balance=False):
    """
    Load the 3DShapes-XOR dataset.
        Note that first need to download the h5 file of the 3dshapes dataset and
        put into "./dataset_storage/3dshapes_xor/3dshapes.h5"

    the basic data loading is credited to https://github.com/deepmind/3d-shapes/blob/master/3dshapes_loading_example.ipynb

    n_samples_for_positive_class = 6000 means to sample 5000 belonging to the positive class

    sample_balance: whether to balance the positive and negative

    The ground truth rule for assigning labels see the filter_condition function below.


    """
    import h5py
    # load dataset
    file_path = './dataset_storage/3dshapes_xor/3dshapes.h5'
    dataset = h5py.File(file_path, 'r')
    print("file loading okay, start processing")
    images = np.asarray(dataset['images'])  # array shape [480000,64,64,3], uint8 in range(256)
    labels = np.asarray(dataset['labels'])  # array shape [480000,6], float64
    image_shape = images.shape[1:]  # [64,64,3]
    label_shape = labels.shape[1:]  # [6]
    n_samples = labels.shape[0]  # 10*10*10*8*4*15=480000


    _FACTORS_IN_ORDER = ['floor_hue', 'wall_hue', 'object_hue', 'scale', 'shape',
                             'orientation']
    _NUM_VALUES_PER_FACTOR = {'floor_hue': 10, 'wall_hue': 10, 'object_hue': 10,
                              'scale': 8, 'shape': 4, 'orientation': 15}

    _NUM_VALUES_PER_FACTOR_Restricted_Choice = {'floor_hue': [0,9], 'wall_hue': [0,9], 'object_hue': [0,3],
                            'scale': [0,7], 'shape': [0,2], 'orientation': [0,14]}

    factor_labels = np.asarray(labels)
    images = np.asarray(images)

    factor_labels_list = labels.tolist()
    def filter_condition(factor_index):
        factor_label = factor_labels_list[factor_index]
        return (factor_label[2]==0.5 and factor_label[4]!=2) or (factor_label[2]!=0.5 and factor_label[4]==2)

    filtered_indices = list(filter(filter_condition,range(factor_labels.shape[0])))

    # the filtered should be 144000 samples
    assert len(filtered_indices)== 3.0/10 * factor_labels.shape[0]

    labels=np.zeros(factor_labels.shape[0])
    labels[filtered_indices]=1
    truth_labels=np.expand_dims(labels,axis=1)
    print("dataset generation okay")

    print("sample subset")
    positive_indices=np.argwhere(truth_labels.squeeze() == 1).squeeze()
    sampled_positive_indices = np.random.choice(positive_indices, n_samples_for_positive_class , replace=False)

    negative_indices=np.argwhere(truth_labels.squeeze() == 0).squeeze()
    if sample_balance is True:
        sampled_negative_indices = np.random.choice(negative_indices, n_samples_for_positive_class , replace=False)
    else:
        sampled_negative_indices = np.random.choice(negative_indices, int(n_samples_for_positive_class / 3 * 7) , replace=False)


    all_sampled_indices=sampled_negative_indices.tolist()+sampled_positive_indices.tolist()

    images,truth_labels,factor_labels = images[all_sampled_indices],truth_labels[all_sampled_indices],factor_labels[all_sampled_indices]


    # transpose the image so that channel comes first
    images=np.transpose(images, (0, 3, 1, 2))
    # yes normalize it please
    images = images / 256.0

    print("dataset prepare okay")
    return images, truth_labels, factor_labels


def get_breast_data():
    print("start loading data")
    from sklearn.datasets import load_breast_cancer
    from sklearn.model_selection import train_test_split
    from neural_dnf.utils import DBEncoder
    import pandas as pd
    data = load_breast_cancer()

    X_df = pd.DataFrame(data['data'], columns=data['feature_names'])
    y_df = pd.DataFrame(data['target'], columns=['class'])
    f_df = pd.DataFrame(zip(data['feature_names'], ['continuous'] * len(data.feature_names)))
    X_train, X_test, y_train, y_test = train_test_split(X_df, y_df, train_size=0.8)
    db_enc = DBEncoder(f_df, discrete=True)
    db_enc.fit(X_train, y_train)
    X_train, y_train = db_enc.transform(X_train, y_train)
    X_test, y_test = db_enc.transform(X_test, y_test)

    N_feature=X_train.shape[-1]
    # N_feature=4
    print(N_feature)
    X_train=X_train[:,:N_feature]
    y_train=y_train[:,1].reshape(-1,1)
    X_test=X_test[:,:N_feature]
    y_test=y_test[:,1].reshape(-1,1)
    return X_train,y_train,X_test,y_test


def get_dsprites_dataloader(batch_size=128,
                            path_to_data='../dsprites-data/dsprites_data.npz'):
    """DSprites dataloader."""
    dsprites_data = DSpritesDataset(path_to_data,
                                    transform=transforms.ToTensor())
    dsprites_loader = DataLoader(dsprites_data, batch_size=batch_size,
                                 shuffle=True)
    return dsprites_loader


def get_chairs_dataloader(batch_size=128,
                          path_to_data='../rendered_chairs_64'):
    """Chairs dataloader. Chairs are center cropped and resized to (64, 64)."""
    all_transforms = transforms.Compose([
        transforms.Grayscale(),
        transforms.ToTensor()
    ])
    chairs_data = datasets.ImageFolder(root=path_to_data,
                                       transform=all_transforms)
    chairs_loader = DataLoader(chairs_data, batch_size=batch_size,
                               shuffle=True)
    return chairs_loader


def get_chairs_test_dataloader(batch_size=62,
                               path_to_data='../rendered_chairs_64_test'):
    """There are 62 pictures of each chair, so get batches of data containing
    one chair per batch."""
    all_transforms = transforms.Compose([
        transforms.Grayscale(),
        transforms.ToTensor()
    ])
    chairs_data = datasets.ImageFolder(root=path_to_data,
                                       transform=all_transforms)
    chairs_loader = DataLoader(chairs_data, batch_size=batch_size,
                               shuffle=False)
    return chairs_loader


def get_celeba_dataloader(batch_size=128, path_to_data='../celeba_64'):
    """CelebA dataloader with (64, 64) images."""
    celeba_data = CelebADataset(path_to_data,
                                transform=transforms.ToTensor())
    celeba_loader = DataLoader(celeba_data, batch_size=batch_size,
                               shuffle=True)
    return celeba_loader


class DSpritesDataset(Dataset):
    """D Sprites dataset."""
    def __init__(self, path_to_data, subsample=1, transform=None):
        """
        Parameters
        ----------
        subsample : int
            Only load every |subsample| number of images.
        """
        self.imgs = np.load(path_to_data)['imgs'][::subsample]
        self.transform = transform

    def __len__(self):
        return len(self.imgs)

    def __getitem__(self, idx):
        # Each image in the dataset has binary values so multiply by 255 to get
        # pixel values
        sample = self.imgs[idx] * 255
        # Add extra dimension to turn shape into (H, W) -> (H, W, C)
        sample = sample.reshape(sample.shape + (1,))

        if self.transform:
            sample = self.transform(sample)
        # Since there are no labels, we just return 0 for the "label" here
        return sample, 0


class CelebADataset(Dataset):
    """CelebA dataset with 64 by 64 images."""
    def __init__(self, path_to_data, subsample=1, transform=None):
        """
        Parameters
        ----------
        subsample : int
            Only load every |subsample| number of images.
        """
        self.img_paths = glob.glob(path_to_data + '/*')[::subsample]
        self.transform = transform

    def __len__(self):
        return len(self.img_paths)

    def __getitem__(self, idx):
        sample_path = self.img_paths[idx]
        sample = imread(sample_path)

        if self.transform:
            sample = self.transform(sample)
        # Since there are no labels, we just return 0 for the "label" here
        return sample, 0
