###############################################################################
# Provides the required elements to implement the Binary-valued DNF layer and the Boat Optimizer in
#
# This file contains the following elements are implemented:
# * DNFLayer
# * DNFLayer_binary_with_latent_weight
# * MLLPLayer
# * MLP
# * Linear Layer
# * SENN_Coefficient_Layer
#
###############################################################################

from typing import TypeVar, Union, Tuple, Optional, Callable

import math
import torch
import torch.nn as nn
import torch.nn.functional as f

from torch import Tensor
from torch.autograd import Function
from torch.optim.optimizer import Optimizer
from torch.optim import Adam

import numpy as np

import pytorch_lightning as pl

THRESHOLD=0.5

################################################################################

# taken from https://github.com/pytorch/pytorch/blob/bfeff1eb8f90aa1ff7e4f6bafe9945ad409e2d97/torch/nn/common_types.pyi

T = TypeVar("T")
_scalar_or_tuple_2_t = Union[T, Tuple[T, T]]
_size_2_t = _scalar_or_tuple_2_t[int]

################################################################################
# Quantizers


class Binarize(Function):
    '''used where binarization of real-valued parameter into binary values are needed. For example the test case of MLLP'''
    clip_value = 0.5

    @staticmethod
    def forward(ctx, inp):
        ctx.save_for_backward(inp)

        # output = inp.sign()
        output = torch.where(inp > Binarize.clip_value, torch.ones_like(inp).type_as(inp), torch.zeros_like(inp).type_as(inp))
        return output

    @staticmethod
    def backward(ctx, grad_output):
        inp: Tensor = ctx.saved_tensors[0]

        # clipped = inp.abs() <= Binarize.clip_value
        clipped = (inp-0.5).abs() <= 0.5

        output = torch.zeros(inp.size()).to(grad_output.device)
        output[clipped] = 1
        output[~clipped] = 0

        return output * grad_output


binarize = Binarize.apply

################################################################################
# binary torch layers

class DNFLayer_binary_with_latent_weight(nn.Module):
    def __init__(self, n, input_dim, output_dim, use_not=False,use_noise=False,noise_init=0.2):

        super(DNFLayer_binary_with_latent_weight, self).__init__()
        self.n = n
        self.use_not = use_not
        self.input_dim = input_dim if not use_not else input_dim * 2
        self.output_dim = output_dim
        self.use_noise=use_noise

        init_value,init_value_W2 = self.get_initialization()
        self.W = nn.Parameter(init_value)
        self.W2 = nn.Parameter(init_value_W2)

        noise_init = noise_init

        self.noise_rate_W = nn.Parameter(noise_init*torch.ones(self.n, self.input_dim))
        self.noise_rate_W2 = nn.Parameter(noise_init*torch.ones(output_dim, self.n))


    def forward(self, x: Tensor,test=False) -> Tensor:
        if self.use_not:
            x = torch.cat((x, 1 - x), dim=1)

        W = binarize(self.W)
        W2 = binarize(self.W2)

        if test is True or self.use_noise is False:
            W = binarize(self.W)
            W2 = binarize(self.W2)
        else:

            noise_for_W = torch.where(self.W > THRESHOLD, -torch.rand(self.W.shape), torch.rand(self.W.shape))

            noise_for_W = torch.where(self.W > THRESHOLD, -torch.randn(self.W.shape).clamp_(0,1), torch.randn(self.W.shape).clamp_(0,1))
            W = W + ( self.noise_rate_W.abs()) * noise_for_W

            noise_for_W2 = torch.where(self.W2 > THRESHOLD, -torch.randn(self.W2.shape).clamp_(0,1), torch.randn(self.W2.shape).clamp_(0,1))
            W2 = W2 + ( self.noise_rate_W2.abs() ) * noise_for_W2

        temp = torch.prod((1 - (1 - x)[:, :, None] * W.t()[None, :, :]), dim=1)
        temp = (1 - torch.prod(1 - temp[:, :, None]* W2.t()[None, :, :], dim=1) )
        return temp

    def get_initialization(self):
        '''
            In order to make the threholded weight sparse, we follow a similar procedure with DNFLayer_Binary.get_initialization()

            Note that we are generating the **latent** weight (before the threshold)

        '''

        rate=0.1
        init_value =  torch.rand(self.n, self.input_dim)
        init_value_W2 = torch.rand(self.output_dim, self.n)
        init_value = torch.where(init_value<0.25,torch.ones_like(init_value),torch.zeros_like(init_value))
        if self.use_not is True:
            input_dim = int(self.input_dim/2)
            init_value = torch.cat( (init_value[:,:input_dim],init_value[:,input_dim:]*(1-init_value[:,:input_dim]) ) ,dim=1)
        init_value = torch.where(init_value>0.5,1-0.5*torch.rand(self.n, self.input_dim),0.5*torch.rand(self.n, self.input_dim))
        init_value_W2 = torch.where(init_value_W2<0.25,torch.ones_like(init_value_W2),torch.zeros_like(init_value_W2))
        init_value_W2 = torch.where(init_value_W2>THRESHOLD,1-0.5*torch.rand(self.output_dim, self.n),0.5*torch.rand(self.output_dim, self.n))
        return init_value,init_value_W2


    def regularization(self):
        regularization = 0
        # regularization += self.lambda_g * torch.mm(self.W2,self.W).mean()
        regularization += self.lambda_g * self.dnf_size()
        return regularization

    def dnf_size(self):
        # self.noise_rate_W.abs().mean()+self.noise_rate_W2.abs().mean()
        actual_dnf_size = torch.mm(binarize(self.W2),binarize(self.W) ).sum()
        return actual_dnf_size

    def clip_params(self):
        self.noise_rate_W.data.clamp_(0,0.1)
        self.noise_rate_W2.data.clamp_(0,0.1)

        self.W.data.clamp_(0,1)
        self.W2.data.clamp_(0,1)

    def test_time_forward(self,x):
        return self.forward(x,test=True)


class MLLPLayer(nn.Module):
    def __init__(self, n, input_dim, output_dim, use_not=False):

        super( MLLPLayer, self).__init__()
        self.n = n
        self.use_not = use_not
        self.input_dim = input_dim if not use_not else input_dim * 2

        self.W = nn.Parameter(0.1 * torch.rand(self.n, self.input_dim))
        self.W2 = nn.Parameter(0.1 * torch.rand(output_dim, self.n))

    def forward(self, x: Tensor,binarized_forward=False) -> Tensor:
        if self.use_not:
            x = torch.cat((x, 1 - x), dim=1)
        if binarized_forward is True:
            W = torch.where(self.W > THRESHOLD, torch.ones_like(self.W), torch.zeros_like(self.W)).type(torch.int)
            W2 = torch.where(self.W2 > THRESHOLD, torch.ones_like(self.W2), torch.zeros_like(self.W2)).type(torch.int)
        else:
            W = self.W
            W2 = self.W2
        temp = torch.prod((1 - (1 - x)[:, :, None] * W.t()[None, :, :]), dim=1)
        temp = (1 - torch.prod(1 - temp[:, :, None]* W2.t()[None, :, :], dim=1) )
        return temp


    def test_time_forward(self,x):
        return self.forward(x,binarized_forward=True)


    def regularization(self):

        regularization = 0

        regularization += self.lambda_g * ( (self.W**2).sum()+ (self.W2**2).sum())
        return regularization

class MLPLayer(nn.Module):
    def __init__(self, n, input_dim, output_dim, use_not=False):

        super(MLPLayer, self).__init__()
        self.n = n
        self.use_not = use_not
        self.input_dim = input_dim if not use_not else input_dim * 2


        n_hid = n
        self.computation = nn.Sequential(
            nn.Linear(self.input_dim, n_hid),
            nn.ReLU(),
            # nn.Linear(n_hid, n_hid),
            # nn.ReLU(),
            nn.Linear(n_hid, output_dim),
            nn.Sigmoid(),
        )
    def forward(self, x: Tensor) -> Tensor:
        if self.use_not:
            x = torch.cat((x, 1 - x), dim=1)
        return self.computation(x)
        # return (1 - torch.prod(1 - temp, dim=1) ).squeeze()

    def test_time_forward(self,x):
        return self.forward(x)


    def regularization(self):
        return 0

class LinearLayer(nn.Module):
    def __init__(self, input_dim, output_dim, use_not=False):

        super(LinearLayer, self).__init__()
        self.use_not = use_not
        self.input_dim = input_dim if not use_not else input_dim * 2

        if output_dim==1:
            self.computation = nn.Sequential(
                nn.Linear(self.input_dim, output_dim),
                nn.Sigmoid(),
            )
        else:
            self.computation = nn.Sequential( nn.Linear(self.input_dim, output_dim), )

    def forward(self, x: Tensor) -> Tensor:
        if self.use_not:
            x = torch.cat((x, 1 - x), dim=1)
        return self.computation(x)
        # return (1 - torch.prod(1 - temp, dim=1) ).squeeze()

    def test_time_forward(self,x):
        return self.forward(x)

    def regularization(self):
        return 0


class SENN_Coefficient_Layer(nn.Module):
    def __init__(self, input_dim, output_dim,raw_input_size=None):
        super(SENN_Coefficient_Layer, self).__init__()
        if raw_input_size is None:
            self.input_dim = input_dim
            self.concept_dim = input_dim
        else:
            self.input_dim = np.prod(raw_input_size)
            self.concept_dim = input_dim
        self.output_dim = output_dim

        self.computation = nn.Sequential(
            nn.Linear(self.input_dim, int(self.input_dim / 10)),
            nn.ReLU(),
            nn.Linear(int(self.input_dim / 10), int(self.input_dim / 50)),
            nn.ReLU(),
            nn.Linear(int(self.input_dim / 50),  self.concept_dim * self.output_dim),
        )

    def forward(self, x: Tensor) -> Tensor:
        x = x.view(x.shape[0],-1)
        result = self.computation(x)
        result = result.view(result.shape[0], self.concept_dim, self.output_dim)
        return result

    def test_time_forward(self,x):
        return self.forward(x)

    def regularization(self):
        return 0

################################################################################
# here starts for the Bop (Helwegen, Koen, et al. 2019)

class MomentumWithThresholdBinaryOptimizer(Optimizer):
    def __init__(
        self,
        binary_params,
        ar = 0.0001,
        threshold = 0.00001,
        continuous_optimizer = None
    ):
        if not 0 < ar < 1:
            raise ValueError(
                "given adaptivity rate {} is invalid; should be in (0, 1) (excluding endpoints)".format(
                    ar
                )
            )
        if threshold < 0:
            raise ValueError(
                "given threshold {} is invalid; should be > 0".format(threshold)
            )

        self.total_weights = {}
        if continuous_optimizer is None:
            self.use_non_binary=False
        else:
            self.use_non_binary=True
            self._adam = continuous_optimizer


        defaults = dict(adaptivity_rate=ar, threshold=threshold)
        super(MomentumWithThresholdBinaryOptimizer, self).__init__(
            binary_params, defaults
        )

    def step(self, closure: Optional[Callable[[], float]] = ..., ar=None):
        if self.use_non_binary is True:
            self._adam.step()

        flips = {None}

        for group in self.param_groups:
            params = group["params"]

            y = group["adaptivity_rate"]
            t = group["threshold"]
            flips = {}

            if ar is not None:
                y = ar

            for param_idx, p in enumerate(params):
                grad = p.grad.data

                state = self.state[p]
                if len(state) == 0:
                    state['step'] = 0
                    state["moving_average"] = torch.zeros_like(p.data, memory_format=torch.preserve_format)

                m = state['moving_average']
                m.mul_((1 - y))
                m.add_(grad.mul(y))
                m_=m

                # p_temp = p.clone()
                # p_temp[p_temp<0.5]=-1
                # p_temp[p_temp>0.5]=1
                p_temp = torch.where(p>0.5, torch.ones_like(p), - torch.ones_like(p))


                mask = (m_.abs() >= t ) * (m_.sign() == p_temp.sign())
                mask = mask.float() * 1
                flips[param_idx] = mask

                p.data.logical_xor_(mask).type(torch.float)

            # print('# of flip',flips)
        return flips,grad

    def zero_grad(self) -> None:
        super().zero_grad()
        if self.use_non_binary is True:
            self._adam.zero_grad()

class ImprovedMomentumWithThresholdBinaryOptimizer(MomentumWithThresholdBinaryOptimizer):
    def __init__(
        self,
        binary_params,
        ar = 0.0001,
        threshold = 0.00001,
        continuous_optimizer=None
    ):
        super(ImprovedMomentumWithThresholdBinaryOptimizer, self).__init__(
            binary_params=binary_params,ar=ar,threshold=threshold,continuous_optimizer=continuous_optimizer
        )

    def step(self, closure: Optional[Callable[[], float]] = ..., ar=None):

        if self.use_non_binary is True:
            self._adam.step()

        flips = {None}

        for group in self.param_groups:
            params = group["params"]

            y = group["adaptivity_rate"]
            t = group["threshold"]
            flips = {}

            if ar is not None:
                y = ar


            grad = torch.zeros(1)

            for param_idx, p in enumerate(params):
                if p.grad is None:
                    continue

                grad = p.grad.data

                state = self.state[p]
                if len(state) == 0:
                    state['step'] = 0
                    # state['step'] = torch.zeros_like(p.data, memory_format=torch.preserve_format)
                    state["moving_average"] = torch.zeros_like(p.data, memory_format=torch.preserve_format)
                    # m = state["moving_average"] = torch.clone(grad).detach()

                    # state['exp_avg'] = torch.zeros_like(p.data, memory_format=torch.preserve_format)
                    # Exponential moving average of squared gradient values
                    # state['moving_average_sq'] = torch.zeros_like(p.data, memory_format=torch.preserve_format)

                m = state['moving_average']
                m.mul_((1 - y))
                m.add_(grad.mul(y))

                s = state['step']
                s += 1
                bias_correction_m = 1 - ar ** s
                m_ = m / bias_correction_m

                # p_temp = p.clone()
                # p_temp[p_temp<0.5]=-1
                # p_temp[p_temp>0.5]=1
                p_temp = torch.where(p>0.5, torch.ones_like(p), - torch.ones_like(p))

                mask = (m_.abs() >= t) * (m_.sign() == p_temp.sign())
                mask = mask.float() * 1
                flips[param_idx] = mask

                p.data.logical_xor_(mask).type(torch.float)

        return flips,grad


class DNFLayer_Binary(pl.LightningModule):
    def __init__(self, n, input_dim, output_dim, use_not=False,use_noise=True,use_zero_initialization=False,noise_init = 0.05, noise_init_W2=None):

        super(DNFLayer_Binary, self).__init__()

        self.lambda_g = 0.0001

        self.n = n
        self.use_not = use_not
        self.input_dim = input_dim if not use_not else input_dim * 2
        self.output_dim = output_dim
        self.use_noise = use_noise

        init_value,init_value_W2 = self.get_initialization(use_zero_initialization=use_zero_initialization)
        self.W = nn.Parameter(init_value)
        self.W2 = nn.Parameter(init_value_W2)

        noise_init = noise_init

        self.noise_rate_W = nn.Parameter(noise_init*torch.ones(self.n, self.input_dim))
        if noise_init_W2 is None:
            noise_init_W2=noise_init
        self.noise_rate_W2 = nn.Parameter(noise_init_W2*torch.ones(output_dim, self.n))



    def forward(self, x: Tensor,test=False) -> Tensor:
        if self.use_not:
            x = torch.cat((x, 1 - x), dim=1)
        if test is True or self.use_noise is False:
            W,W2=self.W,self.W2
        else:
            W,W2 = self.get_noisy_weight()

        self.conj = torch.prod((1 - (1 - x)[:, :, None] * W.t()[None, :, :]), dim=1)
        # self.disj  = 1 - torch.prod(1 - self.conj, dim=1)
        self.disj = (1 - torch.prod(1 - self.conj[:, :, None]* W2.t()[None, :, :], dim=1) )
        return self.disj

    def get_noisy_weight(self):
        noise_for_W = torch.where(self.W > THRESHOLD, -torch.rand(self.W.shape).type_as(self.W), torch.rand(self.W.shape).type_as(self.W))
        W = self.W + (self.noise_rate_W.abs() ) * noise_for_W

        noise_for_W2 = torch.where(self.W2 > THRESHOLD, -torch.rand(self.W2.shape).type_as(self.W), torch.rand(self.W2.shape).type_as(self.W) )
        W2 = self.W2 + ( self.noise_rate_W2.abs() ) * noise_for_W2

        self.noisy_W=W
        self.noisy_W2=W2
        return W,W2

    def adjust_params(self):
        pass
        # TODO: remove this?

    def test_time_forward(self,x):
        return self.forward(x,test=True)

    def clip_params(self):
        self.noise_rate_W.data.clamp_(0,0.5)
        self.noise_rate_W2.data.clamp_(0,0.5)
        return

    def regularization(self):
        regularization = 0
        regularization += self.lambda_g * self.dnf_size()
        return regularization

    def get_initialization(self,use_zero_initialization):
        '''
        Initialize the weights. In order to make the binary matrix sparse, we use a bernoulli rate < 0.5 so that there are more 0 than 1.

        Also, we make sure that a literal cannot co-appear with its negation.

        In fact, with the noise, all-zero-initialization also works. e.g.
            init_value = torch.zeros(self.n, self.input_dim)
            init_value_W2 = torch.zeros(self.output_dim, self.n)

        '''
        if use_zero_initialization is False:
            # how likely it is 1
            rate=0.1
            # rate=0.5
            init_value = 1 * torch.rand(self.n, self.input_dim)
            init_value_W2 = torch.rand(self.output_dim, self.n)
            init_value = torch.where(init_value<rate,torch.ones_like(init_value),torch.zeros_like(init_value))
            if self.use_not is True:
                input_dim = int(self.input_dim/2)
                init_value = torch.cat( (init_value[:,:input_dim],init_value[:,input_dim:]*(1-init_value[:,:input_dim]) ) ,dim=1)
            init_value_W2 = torch.where(init_value_W2<rate,torch.ones_like(init_value_W2),torch.zeros_like(init_value_W2))
        else:
            init_value = torch.zeros(self.n, self.input_dim)
            init_value_W2 = torch.zeros(self.output_dim, self.n)

        return init_value,init_value_W2


    def dnf_size(self):
        # self.noise_rate_W.abs().mean()+self.noise_rate_W2.abs().mean()

        actual_dnf_size = torch.mm(self.W2,self.W).sum()/self.output_dim
        return actual_dnf_size
        # return self.W2.sum()

    def actual_noisy_dnf(self):
        noisy_W,noisy_W2 = self.noisy_W,self.noisy_W2
        temp = torch.unsqueeze(self.W2,axis=2).repeat(1,1,self.input_dim)
        temp = torch.squeeze(temp * self.W,axis=0)
        noise_W = temp*self.noise_rate_W
        noise_W2 = self.W2*self.noise_rate_W2
        return noise_W,noise_W2
