###############################################################################
#
# Provides the first-stage neural network module in PyTorch.


# Some of the code are adopted from https://github.com/dmelis/SENN
###############################################################################


from collections import OrderedDict

import torch
import torch.nn as nn
import torch.nn.functional as F

import torchvision.models as models

import numpy as np

import pytorch_lightning as pl

class DummyFeatureExtractor(nn.Module):
    def __init__(self):
        super(DummyFeatureExtractor, self).__init__()

    def forward(self,x,test_flag=False):
        return x

    def test_time_forward(self, x):
        return self.forward(x)

    def regularization(self):
        return 0

    def freeze(self):
        return

class MLP(nn.Module):
    def __init__(self, input_dim, output_dim, expand_dim):
        super(MLP, self).__init__()
        self.expand_dim = expand_dim
        self.linear = nn.Linear(input_dim, output_dim)
        if self.expand_dim:
            self.linear = nn.Linear(input_dim, expand_dim)
            self.activation = torch.nn.ReLU()
            self.linear2 = nn.Linear(expand_dim, output_dim) #softmax is automatically handled by loss function
            modules = [ self.linear,self.activation, self.linear2 ]
            self.base_computation = nn.Sequential(*modules)
        else:
            modules = [ self.linear]
            self.base_computation = nn.Sequential(*modules)

    def forward(self, x,test_flag=False):
        x = x.view(x.size(0), -1)
        x = self.base_computation(x)
        return x

    def test_time_forward(self, x,test_flag=False):
        return self.forward(x)

    def regularization(self):
        return 0

    def freeze(self):
        for param in self.base_computation.parameters():
            param.requires_grad = False
        return


class PretrainedFeatureBase(nn.Module):
    def __init__(self,model_name='resnet', pretrained=False):
        super(PretrainedFeatureBase, self).__init__()
        print("start constructing")
        if model_name=='resnet':
            pretrained_net = models.resnet34(pretrained=True)
            modules = list(pretrained_net.children())[:-1]
        elif model_name=='vgg':
            pretrained_net = models.vgg16(pretrained=True)
            modules = list(pretrained_net.children())[:-1]
        elif model_name =="inception_v3":
            pretrained_net = models.inception_v3(pretrained=True)
            modules = list(pretrained_net.children())[:-1]
            del modules[15]

        self.base_computation = nn.Sequential(*modules)

    def forward(self,x,test_flag=False):
        x = self.base_computation(x)
        x = x.view(x.size(0), -1)
        return x

    def test_time_forward(self, x):
        return self.forward(x)

    def regularization(self):
        return 0

    def freeze(self):
        for param in self.base_computation.parameters():
            param.requires_grad = False
        return


class ImageConvNet(nn.Module):

    def __init__(self, img_size, latent_dim=0):
        """
        Class which defines model and forward pass.

        Parameters
        ----------
        img_size : tuple of ints
            Size of images. E.g. (1, 32, 32) or (3, 64, 64).

        latent_dim: if not equal to zero, we will use a fc layer

        """
        super(ImageConvNet, self).__init__()

        self.latent_dim = latent_dim

        # Parameters
        self.img_size = img_size
        self.num_pixels = img_size[1] * img_size[2]
        self.hidden_dim = 256  # Hidden dimension of linear layer
        self.reshape = (64, 4, 4)  # Shape required to start transpose convs

        # Intial layer
        encoder_layers = [
            nn.Conv2d(self.img_size[0], 32, (4, 4), stride=2, padding=1),
            nn.BatchNorm2d(32),
            nn.ReLU(),
            nn.Dropout2d(0.25)
        ]
        # Add additional layer if (64, 64) images
        if self.img_size[1:] == (64, 64):
            encoder_layers += [
                nn.Conv2d(32, 32, (4, 4), stride=2, padding=1),
                nn.BatchNorm2d(32),
                nn.ReLU(),
                nn.Dropout2d(0.25)
            ]
        elif self.img_size[1:] == (32, 32):
            # (32, 32) images are supported but do not require an extra layer
            pass
        else:
            raise RuntimeError("{} sized images not supported. Only (None, 32, 32) and (None, 64, 64) supported. Build your own architecture or reshape images!".format(img_size))
        # Add final layers
        encoder_layers += [
            nn.Conv2d(32, 64, (4, 4), stride=2, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(),
            nn.Dropout2d(0.25)
            # nn.Conv2d(64, 64, (4, 4), stride=2, padding=1),
            # nn.BatchNorm2d(64),
            # nn.ReLU(),
            # nn.Linear(64 * 4 * 4, self.hidden_dim),
            # nn.ReLU(),
        ]
        self.fc=nn.Sequential(
                nn.Linear(64*8*8, self.latent_dim),
                nn.ReLU()
                )
        self.computation = nn.Sequential(*encoder_layers)

    def forward(self, x,test_flag=False):
        x = self.computation(x)
        if self.latent_dim!=0:
            x = self.fc(x.view(x.shape[0],-1))
        return x

    def test_time_forward(self,x):
        return self.forward(x,test_flag=True)

    def regularization(self):
        return 0



class ImageConv(nn.Module):

    def __init__(self, img_size, latent_dim=0):
        """
        Class which defines model and forward pass.

        Parameters
        ----------
        img_size : tuple of ints
            Size of images. E.g. (1, 32, 32) or (3, 64, 64).

        latent_dim: if not equal to zero, we will use a fc layer

        """
        super(ImageConv, self).__init__()

        self.latent_dim = latent_dim

        # Parameters
        self.img_size = img_size
        self.num_pixels = img_size[1] * img_size[2]
        self.hidden_dim = 256  # Hidden dimension of linear layer
        self.reshape = (64, 4, 4)  # Shape required to start transpose convs

        # Intial layer
        encoder_layers = [
            nn.Conv2d(self.img_size[0], 32, (4, 4), stride=2, padding=1),
            nn.BatchNorm2d(32),
            nn.ReLU(),
            nn.Dropout2d(0.25)
        ]
        # Add additional layer if (64, 64) images
        if self.img_size[1:] == (64, 64):
            encoder_layers += [
                nn.Conv2d(32, 32, (4, 4), stride=2, padding=1),
                nn.BatchNorm2d(32),
                nn.ReLU(),
                nn.Dropout2d(0.25)
            ]
        elif self.img_size[1:] == (32, 32):
            # (32, 32) images are supported but do not require an extra layer
            pass
        else:
            raise RuntimeError("{} sized images not supported. Only (None, 32, 32) and (None, 64, 64) supported. Build your own architecture or reshape images!".format(img_size))
        # Add final layers
        encoder_layers += [
            nn.Conv2d(32, 64, (4, 4), stride=2, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(),
            nn.Dropout2d(0.25)
            # nn.Conv2d(64, 64, (4, 4), stride=2, padding=1),
            # nn.BatchNorm2d(64),
            # nn.ReLU(),
            # nn.Linear(64 * 4 * 4, self.hidden_dim),
            # nn.ReLU(),
        ]
        self.fc=nn.Linear(64*4*4, self.latent_dim)
        self.computation = nn.Sequential(*encoder_layers)

    def forward(self, x,test_flag=False):
        if test_flag is False:
            x+= torch.randn(x.shape)*0.01
        x = self.computation(x)
        if self.latent_dim!=0:
            x = self.fc(x.view(x.shape[0],-1))
        return x

    def test_time_forward(self,x):
        return self.forward(x,test_flag=True)

    def regularization(self):
        return 0

######################################################################

def kl_normal_loss(mean, logvar):
    """
    Calculates the KL divergence between a normal distribution
    with diagonal covariance and a unit normal distribution.
    Parameters
    ----------
    mean : torch.Tensor
        Mean of the normal distribution. Shape (batch_size, latent_dim) where
        D is dimension of distribution.
    logvar : torch.Tensor
        Diagonal log variance of the normal distribution. Shape (batch_size,
        latent_dim)
    storer : dict
        Dictionary in which to store important variables for vizualisation.
    """
    latent_dim = mean.size(1)
    # batch mean of kl for each latent dimension
    latent_kl = 0.5 * (-1 - logvar + mean.pow(2) + logvar.exp()).mean(dim=0)
    total_kl = latent_kl.sum()

    return total_kl

def linear_annealing(init, fin, step, annealing_steps=5000):
    """Linear annealing of a parameter."""
    if annealing_steps == 0:
        return fin
    assert fin > init
    delta = fin - init
    annealed = min(init + delta * step / annealing_steps, fin)
    return annealed

######################################################################

class ImageConvVAE(pl.LightningModule):
    '''
    architecture and code adopted from https://github.com/Schlumberger/joint-vae
    and https://github.com/YannDubs/disentangling-vae
    '''

    def __init__(self, img_size, latent_dim):
        """
        Class which defines model and forward pass.

        Parameters
        ----------
        img_size : tuple of ints
            Size of images. E.g. (1, 32, 32) or (3, 64, 64).

        latent_dim : int
        """
        super(ImageConvVAE, self).__init__()

        # parameters related to the beta-VAE
        self.gamma=100.
        self.C_init = 0.
        self.C_fin = 20.

        # self.steps_anneal = 5000
        self.steps_anneal = 20000

        # Parameters
        self.img_size = img_size
        self.latent_dim = latent_dim
        self.num_pixels = img_size[1] * img_size[2]
        self.hidden_dim = 256  # Hidden dimension of linear layer
        self.reshape = (64, 4, 4)  # Shape required to start transpose convs

        # Define encoder layers
        # Intial layer
        encoder_layers = [
            nn.Conv2d(self.img_size[0], 32, (4, 4), stride=2, padding=1),
            nn.BatchNorm2d(32),
            nn.ReLU()
        ]
        # Add additional layer if (64, 64) images
        if self.img_size[1:] == (64, 64):
            encoder_layers += [
                nn.Conv2d(32, 32, (4, 4), stride=2, padding=1),
                nn.BatchNorm2d(32),
                nn.ReLU()
            ]
        elif self.img_size[1:] == (32, 32):
            # (32, 32) images are supported but do not require an extra layer
            pass
        else:
            raise RuntimeError("{} sized images not supported. Only (None, 32, 32) and (None, 64, 64) supported. Build your own architecture or reshape images!".format(img_size))
        # Add final layers
        encoder_layers += [
            nn.Conv2d(32, 64, (4, 4), stride=2, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(),
            nn.Conv2d(64, 64, (4, 4), stride=2, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU()
        ]

        # Define encoder
        self.img_to_features = nn.Sequential(*encoder_layers)

        # Map encoded features into a hidden vector which will be used to
        # encode parameters of the latent distribution
        self.features_to_latent = nn.Sequential(
            nn.Linear(64 * 4 * 4, self.hidden_dim),
            nn.ReLU(),
            nn.Linear(self.hidden_dim, self.latent_dim*2)
        )

        # Map latent samples to features to be used by generative model
        self.latent_to_features = nn.Sequential(
            nn.Linear(self.latent_dim, self.hidden_dim),
            nn.ReLU(),
            nn.Linear(self.hidden_dim, 64 * 4 * 4),
            nn.ReLU()
        )

        # Define decoder
        decoder_layers = []

        # Additional decoding layer for (64, 64) images
        if self.img_size[1:] == (64, 64):
            decoder_layers += [
                nn.ConvTranspose2d(64, 64, (4, 4), stride=2, padding=1),
                nn.BatchNorm2d(64),
                nn.ReLU()
            ]

        decoder_layers += [
            nn.ConvTranspose2d(64, 32, (4, 4), stride=2, padding=1),
            nn.BatchNorm2d(32),
            nn.ReLU(),
            nn.ConvTranspose2d(32, 32, (4, 4), stride=2, padding=1),
            nn.BatchNorm2d(32),
            nn.ReLU(),
            nn.ConvTranspose2d(32, self.img_size[0], (4, 4), stride=2, padding=1),
            nn.Sigmoid()
        ]

        # Define decoder
        self.features_to_img = nn.Sequential(*decoder_layers)

    def encode(self, x):
        """
        Encodes an image into parameters of a latent distribution defined in
        self.latent_spec.

        Parameters
        ----------
        x : torch.Tensor
            Batch of data, shape (N, C, H, W)
        """
        batch_size = x.size()[0]

        # Encode image to latent features
        features = self.img_to_features(x)
        latent_representation = self.features_to_latent(features.view(batch_size, -1))
        return latent_representation

    def decode(self, latent_sample):
        """
        Decodes sample from latent distribution into an image.

        Parameters
        ----------
        latent_sample : torch.Tensor
            Sample from latent distribution. Shape (N, L) where L is dimension
            of latent distribution.
        """
        features = self.latent_to_features(latent_sample)
        return self.features_to_img(features.view(-1, *self.reshape))

    def reparameterize(self, mean, logvar, test_flag=False):
        """
        Samples from a normal distribution using the reparameterization trick.
        Parameters
        ----------
        mean : torch.Tensor
            Mean of the normal distribution. Shape (batch_size, latent_dim)
        logvar : torch.Tensor
            Diagonal log variance of the normal distribution. Shape (batch_size,
            latent_dim)
        """
        if test_flag is False:
            std = torch.exp(0.5 * logvar)
            eps = torch.randn_like(std)
            return mean + std * eps
        else:
            # Reconstruction mode
            return mean

    def forward(self, x,test_flag=False):
        """
        Forward pass of model.

        Parameters
        ----------
        x : torch.Tensor
            Batch of data. Shape (N, C, H, W)
        """
        mu_logvar = self.encode(x)
        mu, logvar = mu_logvar.view(-1, self.latent_dim, 2).unbind(-1)

        latent_representation = self.reparameterize(mu,logvar,test_flag=test_flag)
        # latent_representation = self.reparameterize(mu,logvar)

        self.latent_dist = (mu,logvar)
        self.latent_sample = latent_representation
        self.reconstructed = self.decode(latent_representation)

        return latent_representation

    def test_time_forward(self,x):
        return self.forward(x,test_flag=True)

    def loss(self,x,global_train_step,test_flag=False):
        # reconstruction error
        # rec_loss=F.mse_loss(self.reconstructed,x)
        rec_loss = F.mse_loss(self.reconstructed, x, reduction="sum")

        increasing_capacity = (linear_annealing(self.C_init, self.C_fin, global_train_step, self.steps_anneal) if test_flag is False else self.C_fin)
        # increasing_capacity = 0

        # KL-loss
        kl_loss = kl_normal_loss(*self.latent_dist)


        regularization_loss =  self.gamma * (kl_loss - increasing_capacity).abs()

        return rec_loss,regularization_loss

    def train_dataloader(self):
        return DataLoader(MNIST(os.getcwd(), train=True, download=True,
                          transform=transforms.ToTensor()), batch_size=32)


    def training_step(self, batch, batch_idx):
        x, y = batch

        latent_representation = self.forward(x)

        rec_loss,regularization_loss = self.loss(x,self.global_step)
        train_loss = rec_loss + regularization_loss

        logger_logs = {
            "train_loss": train_loss,
            'reconstruction loss':rec_loss,
            'KL_loss':regularization_loss}

        # if hasattr(self,'logger') and self.use_logger:

        # loss is strictly required
        output = OrderedDict(
            {
                "loss": train_loss,
                "progress_bar": {"train_loss": train_loss,
                                'reconstruction loss':rec_loss
                                },
                "log": logger_logs,
            }
        )

        return output

    def validation_step(self, batch, batch_idx):
        # OPTIONAL
        x, y = batch
        latent_representation = self.test_time_forward(x)

        rec_loss,regularization_loss = self.loss(x,self.global_step)
        val_loss = rec_loss + regularization_loss

        logger_logs = {
            "val_loss": val_loss,
            'reconstruction loss':rec_loss,
            'KL_loss':regularization_loss}
        # validation metrics for monitoring

        return OrderedDict(
            {
            "val_loss": val_loss.clone().detach(),
            "log": logger_logs
            }
        )

    def configure_optimizers(self):
        # initial_lr = 1e-05
        initial_lr = 0.001

        optimizer = torch.optim.Adam(self.parameters(), lr=initial_lr)

        scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, gamma=0.95)
        return [optimizer],[scheduler]
        # return optimizer
