import os
from collections import OrderedDict



import torch
import numpy as np
import torch.nn as nn
from torch.nn import functional as F
from torch.utils.data import DataLoader
from torchvision.datasets import MNIST
import torchvision.transforms as transforms

import pytorch_lightning as pl

from .phi import ProgressiveBinarization

THRESHOLD = 0.5

class TwoStageModelBase(pl.LightningModule):
    """
    the base class for the base model
    """

    def __init__(self,
                        DNF_dim_list,
                        phi=None,
                        use_not=False,
                        loss_weight_list=[0,1],
                        use_logger=True,
                        loss="MSE",
                        hparams=None):
        """
        Parameters
        ----------
        DNF_dim_list : list
            A list specifies the number of nodes (neurons) of all the layers in NeuralDNF from bottom to top. DNF_dim_list[0]
            should be the dimensionality of the input data and DNF_dim_list[1] should be the number of class labels.
        device : torch.device
            Run on which device.
        use_not : bool
            Whether use the NOT (~) operator in logical rules.
        loss_weight_list: indicate the loss (regularization) for the first stage and the second stage. Note that in this paper we do not utilize any loss (regularization) for the first stage so it is zero. But in the future it is possible to do that.
        loss="MSE" or "BCE", mean square error or binary cross entropy
        """

        super(TwoStageModelBase, self).__init__()

        self.DNF_dim_list = DNF_dim_list
        self.output_dim = DNF_dim_list[-1]
        self.use_not = use_not
        self.use_logger = use_logger
        self.loss_weight_list = loss_weight_list
        self.loss=loss

        if phi is None:
            from .phi import DummyPhiModule
            phi = DummyPhiModule()
            # from .phi import image_Conv
            # n_concept_predicate=20
            # phi=image_Conv(n_concept_predicate)
        self.system_1 = phi
        self.add_module('system_1', self.system_1)

    def forward(self, x):
        x = self.system_1(x)
        x = self.system_2(x)
        return x

    def test_time_forward(self,x):
        x = self.system_1.test_time_forward(x)
        return self.system_2.test_time_forward(x)

    def loss_function(self,y_hat,y):
        '''returns the loss of the second stage'''
        batch_size = y.size()[0]
        if self.loss == "MSE":
            loss =  F.mse_loss(y_hat,y,reduction = 'sum') / batch_size
        elif self.loss == "BCE":
            # print(F.binary_cross_entropy(y_hat,y,reduction = 'sum') / y.size()[0])
            if self.output_dim==1:
                loss = F.binary_cross_entropy(y_hat,y,reduction = 'sum') / batch_size
            else:
                _, labels = y.max(dim=1)
                loss = nn.CrossEntropyLoss()(y_hat, labels)

        loss += self.system_2.regularization()

        return loss

    def train_dataloader(self):
        pass

    def configure_optimizers(self):
        pass

    def val_dataloader(self):
        # return self.train_dataloader()
        pass




    def training_step(self, batch, batch_idx):

        if len(batch)==3:
            x, y, c = batch
        else:
            x, y = batch

        y_hat = self.forward(x)

        if self.output_dim==1:
            labels_hat = torch.where(y_hat > THRESHOLD, torch.ones_like(y_hat), torch.zeros_like(y_hat)).type(torch.int)
            train_acc = torch.sum(y == labels_hat).item() / (y.shape[0] * 1.0)
        else:
            labels_ = torch.argmax(y_hat, dim=1)
            train_acc = torch.sum(y == labels_).item() / (y.shape[0] * 1.0)
            y = F.one_hot(y, self.output_dim).float()

        if len(batch)!=3:
            regularization_loss = self.system_1.regularization(x,global_train_step=self.global_step)
            system_1_loss = regularization_loss
        else:
            regularization_loss = self.system_1.regularization(x,global_train_step=self.global_step,concept_annotation=c)
            system_1_loss = regularization_loss

        system_2_loss = self.loss_function(y_hat, y)
        loss_weight_1,loss_weight_2=self.loss_weight_list
        train_loss = loss_weight_1*system_1_loss + loss_weight_2*system_2_loss

        logger_logs = {"train_acc": train_acc,
            "train_loss": train_loss,
            'system_1_loss':system_1_loss, 'system_2_loss':system_2_loss,}

        # loss is strictly required
        output = OrderedDict(
            {
                "loss": train_loss,
                "system_1_loss":system_1_loss,
                "progress_bar": {"train_acc": train_acc},
                "log": logger_logs,
            }
        )

        return output


    def validation_step(self, batch, batch_idx):
        # OPTIONAL
        if len(batch)==3:
            x, y, c = batch
        else:
            x, y = batch

        y_hat = self.test_time_forward(x)

        if self.output_dim==1:
            labels_hat = torch.where(y_hat > THRESHOLD, torch.ones_like(y_hat), torch.zeros_like(y_hat)).type(torch.int)
            val_acc = torch.sum(y == labels_hat).item() / (y.shape[0] * 1.0)
        else:
            labels_ = torch.argmax(y_hat, dim=1)
            val_acc = torch.sum(y == labels_).item() / (y.shape[0] * 1.0)
            y = F.one_hot(y, num_classes=self.output_dim).float()


        # validation metrics for monitoring
        val_loss = F.mse_loss(y_hat, y,reduction = 'sum') / y.size()[0]
        logger_logs = {"val_acc": val_acc, "val_loss": val_loss}

        return OrderedDict(
            {
            "val_loss": val_loss.clone().detach(),
            "val_acc": torch.tensor(val_acc),
            "log": logger_logs
            }
        )

    def validation_epoch_end(self, outputs):
        avg_loss = torch.stack([x['val_loss'] for x in outputs]).mean()
        avg_acc = torch.stack([x['val_acc'] for x in outputs]).mean()
        tensorboard_logs = {'val_loss': avg_loss,'val_acc': avg_acc}
        return {'val_loss': avg_loss, 'val_acc': avg_acc,'log': tensorboard_logs}

# TODO: Add the randomized binarization
class NeuralDNF(TwoStageModelBase):

    def __init__(self,
                        DNF_dim_list,
                        phi=None,
                        use_not=False,
                        loss_weight_list=[0,1],
                        use_noise=True,
                        use_logger=True,
                        loss="MSE",
                        hparams=None):
        super(NeuralDNF, self).__init__(DNF_dim_list, phi, use_not, loss_weight_list, use_logger, loss, hparams)


        from .core import DNFLayer_binary_with_latent_weight
        self.system_2 = DNFLayer_binary_with_latent_weight(DNF_dim_list[1],
                                input_dim=DNF_dim_list[0],
                                output_dim=DNF_dim_list[-1],
                                use_not=use_not,
                                use_noise=use_noise)

        self.add_module('system_2', self.system_2)

        self.set_parameter()

    def set_parameter(self):
        self.adam_lr=0.001
        self.weight_decay=0.0


    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.adam_lr, weight_decay=self.weight_decay)

        # scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, gamma=0.95)
        scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=50, gamma=0.75)
        return [optimizer],[scheduler]
        # return optimizer

    def optimizer_step(
        self, current_epoch, batch_nb, optimizer, optimizer_i, second_order_closure=None
    ):
        optimizer.step()
        self.system_2.clip_params()
        optimizer.zero_grad()

        if hasattr(self,'logger'):
            self.logger.experiment.add_scalar("noise rate W", self.system_2.noise_rate_W.mean(), self.global_step)
            self.logger.experiment.add_scalar("noise rate W2", self.system_2.noise_rate_W2.mean(), self.global_step)

            self.logger.experiment.add_scalar("DNF actual size", self.system_2.dnf_size(), self.global_step)

class NeuralDNF_BOAT(TwoStageModelBase):
    """
    The Neural Disjunctive Normal form
    """

    def __init__(self,
                        DNF_dim_list,
                        phi=None,
                        use_not=False,
                        loss_weight_list=[0,1],
                        use_noise=True,
                        use_logger=True,
                        use_zero_initialization=False,
                        loss="MSE",
                        noise_init=0.2,
                        noise_init_W2=None,
                        hparams=None):


        super(NeuralDNF_BOAT, self).__init__(DNF_dim_list, phi, use_not, loss_weight_list, use_logger, loss, hparams)


        # from .core import DNFLayer_Binary
        # self.system_2 = DNFLayer_Binary(DNF_dim_list[1],
        #                         input_dim=DNF_dim_list[0],
        #                         output_dim=DNF_dim_list[-1],
        #                         use_not=use_not,
        #                         use_zero_initialization=use_zero_initialization,
        #                         use_noise=use_noise,
        #                         noise_init_W2=noise_init_W2,
        #                         noise_init=noise_init)
        #                         self.add_module('system_2', self.system_2)
        from .core import DNFLayer_Binary
        if DNF_dim_list[-1]==1:
            self.system_2 = DNFLayer_Binary(DNF_dim_list[1],
                                    input_dim=DNF_dim_list[0],
                                    output_dim=DNF_dim_list[-1],
                                    use_not=use_not,
                                    use_zero_initialization=use_zero_initialization,
                                    use_noise=use_noise,
                                    noise_init_W2=noise_init_W2,
                                noise_init=noise_init)
            self.add_module('system_2', self.system_2)
        else:
            self.system_2 = []
            for idx in range(DNF_dim_list[-1]):

                layer = DNFLayer_Binary(DNF_dim_list[1],
                                input_dim=DNF_dim_list[0],
                                output_dim=1,
                                use_not=use_not,
                                use_zero_initialization=use_zero_initialization,
                                use_noise=use_noise,
                                noise_init_W2=noise_init_W2,
                                noise_init=noise_init)
                self.system_2.append(layer)
                self.add_module('system_2_'+str(idx), layer)

        self.output_dim = DNF_dim_list[-1]

        self.set_parameter()


        self.total_sum_flip=[0,0]

    def forward(self, x):
        x = self.system_1(x)
        if self.output_dim==1:
            x = self.system_2(x)
        else:
            temp = [ s(x) for s in self.system_2 ]
            x = torch.cat(temp,1)
        return x

    def test_time_forward(self,x):
        x = self.system_1.test_time_forward(x)
        if self.output_dim==1:
            return self.system_2.test_time_forward(x)
        else:
            temp = [ s.test_time_forward(x) for s in self.system_2 ]
            return torch.cat(temp,1)

    def loss_function(self,y_hat,y):
        '''returns the loss of the second stage'''
        batch_size = y.size()[0]
        if self.loss == "MSE":
            loss =  F.mse_loss(y_hat,y,reduction = 'sum') / batch_size
        elif self.loss == "BCE":
            # print(F.binary_cross_entropy(y_hat,y,reduction = 'sum') / y.size()[0])

            loss = F.binary_cross_entropy(y_hat,y,reduction = 'sum') / batch_size

            # if self.output_dim==1:
            #     loss = F.binary_cross_entropy(y_hat,y,reduction = 'sum') / batch_size
            # else:
            #     _, labels = y.max(dim=1)
            #     # loss = nn.CrossEntropyLoss(reduction='sum')(y_hat, labels) / batch_size * self.output_dim
            #     loss = nn.CrossEntropyLoss(reduction='sum')(y_hat, labels) / batch_size

        if self.output_dim==1:
            loss += self.system_2.regularization()
        else:
            for s in self.system_2:
                loss += s.regularization()

        return loss


    def set_parameter(self):
        self.ar = 0.00001
        self.t = 0.000001
        # self.t = 0.00001


        self.adam_lr = 0.01 # learning rate for the Phi hyperplanes
        self.adam_decay_n_epochs = 1
        # self.adam_lr_decay_rate=0.95
        self.adam_lr_decay_rate=1

        self.adam_lr_for_noise = 0.001 # learning rate for the noise rate (the learnable temperature)

        self.weight_decay=0.0
        self.decay_n_epochs = 1
        # self.ar_decay_rate = 0.95
        self.ar_decay_rate = 1


    def configure_optimizers(self):
        if self.output_dim==1:
            continuous_parameters=[self.system_2.noise_rate_W,self.system_2.noise_rate_W2]
        else:
            continuous_parameters=[]
            for s in self.system_2:
                continuous_parameters.append(s.noise_rate_W)
                continuous_parameters.append(s.noise_rate_W2)

        if len( list(self.system_1.parameters() ) ) >0:
            continuous_optimizer = torch.optim.Adam(
                    [
                     {'params': self.system_1.parameters(), 'lr':self.adam_lr},
                     {'params': continuous_parameters, 'lr': self.adam_lr_for_noise}
                 ], lr=self.adam_lr, weight_decay=self.weight_decay)
        else:
            continuous_optimizer = torch.optim.Adam(
                 continuous_parameters, lr=self.adam_lr_for_noise, weight_decay=self.weight_decay)


        if self.output_dim==1:
            binary_parameters = [self.system_2.W,self.system_2.W2]
        else:
            binary_parameters=[]
            for s in self.system_2:
                binary_parameters.append(s.W)
                binary_parameters.append(s.W2)

        from .core import ImprovedMomentumWithThresholdBinaryOptimizer
        binary_optimizer = ImprovedMomentumWithThresholdBinaryOptimizer(
            binary_params=binary_parameters,
            ar=self.ar,
            threshold=self.t,
            continuous_optimizer=continuous_optimizer
        )

        return binary_optimizer

    def optimizer_step(
            self, current_epoch, batch_nb, optimizer, optimizer_i, second_order_closure=None
        ):
            """
            Adaptivity rate decay
            """
            # Decay every n epochs
            if (
                batch_nb == 0
                and current_epoch % self.adam_decay_n_epochs == 0
                and current_epoch != 0
            ):
                self.adjust_adam_optimizer_params(optimizer)
            if (
                batch_nb == 0
                and current_epoch % self.decay_n_epochs == 0
                and current_epoch != 0
            ):
                self.adjust_bop_optimizer_params(optimizer)


            # update params - optimizer step
            flips_curr_step,grad = optimizer.step(ar=self.ar)


            # TODO: see this experimental logger
            if hasattr(self,'logger') and self.use_logger:
                # sum_flips = [f.type(torch.int).sum().item() for f in flips_curr_step.values()]
                #
                # if sum(sum_flips)>0:
                #     for idx in range(2):
                #         self.total_sum_flip[idx]+=sum_flips[idx]
                # scalars
                # for idx in range(2):
                #     self.logger.experiment.add_scalar("total flips_W_"+str(idx+1),self.total_sum_flip[idx],self.global_step)
                # self.logger.experiment.add_scalar("total flips_all",self.total_sum_flip[0]+self.total_sum_flip[1],self.global_step)
                # self.logger.experiment.add_scalar("l1-norm",torch.sum(self.system_2.W.detach()).numpy(),self.global_step)

                # self.logger.experiment.add_scalar("ar", self.ar, self.global_step)
                self.logger.experiment.add_scalar("grad of g", torch.sum(grad), self.global_step)
                if self.output_dim==1:
                    self.logger.experiment.add_scalar("noise rate W", self.system_2.noise_rate_W.mean(), self.global_step)
                    self.logger.experiment.add_scalar("noise rate W2", self.system_2.noise_rate_W2.mean(), self.global_step)
                    self.logger.experiment.add_scalar("DNF actual size", self.system_2.dnf_size(), self.global_step)
                    self.logger.experiment.add_scalar("number of rules", self.system_2.W2.sum(), self.global_step)
                else:
                    self.logger.experiment.add_scalar("noise rate W", torch.mean(torch.FloatTensor([s.noise_rate_W.mean() for s in self.system_2])), self.global_step)
                    self.logger.experiment.add_scalar("noise rate W2", torch.mean(torch.FloatTensor([s.noise_rate_W2.mean() for s in self.system_2])), self.global_step)
                    self.logger.experiment.add_scalar("DNF actual size",
                    torch.mean(torch.FloatTensor([s.dnf_size() for s in self.system_2])), self.global_step)
                    self.logger.experiment.add_scalar("number of rules", torch.mean(torch.FloatTensor([s.W2.sum() for s in self.system_2])), self.global_step)

                # self.logger.experiment.add_scalar("noise rate W corresponding to actual selected", (self.system_2.noise_rate_W*self.system_2.W).mean(), self.global_step)
                # self.logger.experiment.add_scalar("noise rate W2 corresponding to actual selected", (self.system_2.noise_rate_W2*self.system_2.W2).mean(), self.global_step)

                # histograms

                # S = self.system_2.W2
                # s = torch.Tensor()
                # s=torch.Tensor(S.size()[0],1)
                # for c in range(n_class):
                #     s[c]=sum(S[c])
                # self.logger.experiment.add_histogram("h1",rule_per_class, self.global_step)


                # temp1,temp2=self.system_2.actual_noisy_dnf()
                # if len(temp1.detach().numpy())!=0:
                #     self.logger.experiment.add_histogram("h1",temp1, self.global_step)
                # if len(temp2.detach().numpy())!=0:
                #     self.logger.experiment.add_histogram("h2",temp2, self.global_step)
                # self.logger.experiment.add_histogram("W",self.system_2.W, self.global_step)
                # self.logger.experiment.add_histogram("W2",self.system_2.W2, self.global_step)

            if self.output_dim==1:
                self.system_2.clip_params()
            else:
                for s in self.system_2:
                    s.clip_params()
            optimizer.zero_grad()


    def adjust_bop_optimizer_params(self,optimizer):
        self.ar *= self.ar_decay_rate
        return

    def adjust_adam_optimizer_params(self,optimizer):
        for param_group in optimizer._adam.param_groups:
            param_group['lr'] *= self.adam_lr_decay_rate
        return





class MLLP(NeuralDNF):

    def __init__(self,
                            DNF_dim_list,
                            phi=None,
                            use_not=False,
                            loss_weight_list=[0,1],
                            use_logger=True,
                            loss="MSE",
                            hparams=None):
        super(MLLP, self).__init__(DNF_dim_list, phi, use_not, loss_weight_list, False, use_logger, loss, hparams)

        from .core import MLLPLayer
        self.system_2 = MLLPLayer(DNF_dim_list[1],
                                input_dim=DNF_dim_list[0],
                                output_dim=DNF_dim_list[-1],
                                use_not=use_not)

        self.add_module('system_2', self.system_2)

        self.adam_lr=0.001 # in Wang et al 2019, this is 0.005, in  Payani and Fekri 2019  this is 0.001. A larger learning rate learns faster.

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.adam_lr, weight_decay=self.weight_decay)

        scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, gamma=0.95)
        # return [optimizer],[scheduler]
        return optimizer

    def test_time_forward(self,x):
        return self.system_2.test_time_forward(x)

    def optimizer_step(
        self, current_epoch, batch_nb, optimizer, optimizer_i, second_order_closure=None
    ):
        optimizer.step()

        """Clip the weights into the range [0, 1]."""
        for param in self.system_2.parameters():
            param.data.clamp_(0, 1)

        optimizer.zero_grad()


class LinearModel(NeuralDNF):
    def __init__(self,
                                DNF_dim_list,
                                phi=None,
                                use_not=False,
                                loss_weight_list=[0,1],
                                use_logger=True,
                                loss="MSE",
                                hparams=None):
        super(LinearModel, self).__init__(DNF_dim_list, phi, use_not, loss_weight_list, False, use_logger, loss, hparams)

        from .core import LinearLayer
        self.system_2 = LinearLayer(
                                input_dim=DNF_dim_list[0],
                                output_dim=DNF_dim_list[-1],
                                use_not=use_not)
        self.add_module('system_2', self.system_2)

    def configure_optimizers(self):
        # return [torch.optim.Adam(self.parameters(), lr=self.adam_lr, weight_decay=self.weight_decay)],[scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=50, gamma=0.75)]


        self.optimizer = torch.optim.Adam(self.parameters(), lr=self.adam_lr, weight_decay=self.weight_decay)
        # self.optimizer = torch.optim.SGD(self.parameters(), lr=self.adam_lr, momentum=0.9, weight_decay=self.weight_decay)

        self.scheduler = torch.optim.lr_scheduler.StepLR(self.optimizer, step_size=1000, gamma=0.5)
        return self.optimizer

    def optimizer_step(
        self, current_epoch, batch_nb, optimizer, optimizer_i, second_order_closure=None
    ):
        optimizer.step()
        optimizer.zero_grad()

        self.scheduler.step()



class MLP(LinearModel):
    def __init__(self,
                            DNF_dim_list,
                            phi=None,
                            use_not=False,
                            loss_weight_list=[0,1],
                            use_logger=True,
                            loss="MSE",
                            hparams=None):
        super(MLP, self).__init__(DNF_dim_list, phi, use_not, loss_weight_list, use_logger, loss, hparams)

        from .core import MLPLayer
        self.system_2 = MLPLayer(DNF_dim_list[1],
                                input_dim=DNF_dim_list[0],
                                output_dim=DNF_dim_list[-1],
                                use_not=use_not)


        self.add_module('system_2', self.system_2)


class SENNModel(LinearModel):
    def __init__(self,
                            DNF_dim_list,
                            phi=None,
                            use_not=False,
                            loss_weight_list=[0,1],
                            use_logger=True,
                            raw_input_size=None,
                            loss="MSE",
                            hparams=None):
        super(SENNModel, self).__init__(DNF_dim_list, phi, use_not, loss_weight_list, use_logger, loss, hparams)

        from .core import SENN_Coefficient_Layer
        self.system_2 = SENN_Coefficient_Layer(
                                input_dim=DNF_dim_list[0],
                                output_dim=DNF_dim_list[-1],
                                raw_input_size=raw_input_size)
        self.output_dim=DNF_dim_list[-1]
        self.add_module('system_2', self.system_2)

    def forward(self, x, test=False):
        '''
        note that now the system_2 only provides the coefficients
        '''
        phi_x = self.system_1(x)
        coefficient = self.system_2(x)

        phi_x = phi_x.view(phi_x.shape[0],phi_x.shape[1],1)
        # phi_x = phi_x.repeat(1,1,self.output_dim)

        result = torch.bmm(coefficient.transpose(1,2), phi_x).squeeze(dim=-1)
        if self.output_dim==1:
            result = torch.sigmoid(result)
        else:
            result =  torch.log_softmax(result, dim = 1)
        return result

    def test_time_forward(self,x):
        # return self.forward(x)
        phi_x = self.system_1.test_time_forward(x)
        coefficient = self.system_2.test_time_forward(x)

        phi_x = phi_x.view(phi_x.shape[0],phi_x.shape[1],1)
        # phi_x = phi_x.repeat(1,1,self.output_dim)

        result = torch.bmm(coefficient.transpose(1,2), phi_x).squeeze(dim=-1)
        if self.output_dim==1:
            result = torch.sigmoid(result)
        else:
            result =  torch.log_softmax(result, dim = 1)
        return result


class NeuralDNF_BOAT_fixed_temperature(NeuralDNF_BOAT):
    """
    The Neural Disjunctive Normal form
    """

    def __init__(self,
                        DNF_dim_list,
                        phi=None,
                        use_not=False,
                        loss_weight_list=[0,1],
                        use_noise=True,
                        use_logger=True,
                        use_zero_initialization=False,
                        loss="MSE",
                        hparams=None):


        super(NeuralDNF_BOAT_fixed_temperature, self).__init__(DNF_dim_list, phi, use_not, loss_weight_list,use_noise, use_logger, loss, hparams)


        from .core import DNFLayer_Binary
        self.system_2 = DNFLayer_Binary(DNF_dim_list[1],
                                input_dim=DNF_dim_list[0],
                                output_dim=DNF_dim_list[-1],
                                use_not=use_not,
                                use_zero_initialization=use_zero_initialization,
                                use_noise=use_noise)

        for param in [self.system_2.noise_rate_W,self.system_2.noise_rate_W2]:
            param.requires_grad = False

        self.set_parameter()
        self.add_module('system_2', self.system_2)

        self.total_sum_flip=[0,0]

    def adjust_bop_optimizer_params(self,optimizer):
        self.ar *= self.ar_decay_rate
        with torch.no_grad():
            self.system_2.noise_rate_W.data *= 0.9
            self.system_2.noise_rate_W2.data *= 0.9
        return
