###############################################################################
#
# Provides the first-stage neural network module in PyTorch.


# Some of the code are adopted from https://github.com/dmelis/SENN
###############################################################################


from collections import OrderedDict

import math
import torch
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from torch.autograd import Function,Variable
import numpy as np

from .core import Binarize

import pytorch_lightning as pl



######################################################################
def ImprovedSemanticHashing(x,test_flag=False,device='cpu'):
    if test_flag is False:
        # noise = Variable(input.data.new(input.size()).normal_(0, 1))
        noise = torch.randn(x.size())
        noise = noise.type_as(x)
        x = x + noise

    x = torch.sigmoid(x)

    x_binarized = Binarize.apply(x)
    # return input: note in train time, half of the samples are binarized
    if test_flag is True:
        return x_binarized
    else:
        # half-time use real valued half time use binarized
        condition_random = torch.rand(x.shape[0],1).repeat([1,x.shape[1]]).type_as(x)
        return torch.where(condition_random > 0.5, x, x_binarized)


# TODO: remove it?
class GlobalPooling2D(nn.Module):
    def __init__(self):
        super(GlobalPooling2D,self).__init__()

    def forward(self,x):
        x = F.max_pool2d(x, kernel_size=x.size()[2:])
        # x = F.avg_pool2d(x, kernel_size=x.size()[2:])
        return x.view(x.size()[0],-1)

def ProgressiveBinarization(x,global_train_step,test_flag):

    def linear_annealing(init, fin, step, annealing_steps=5000):
        """Linear annealing of a parameter."""
        if annealing_steps == 0:
            return fin
        assert fin > init
        delta = fin - init
        annealed = min(init + delta * step / annealing_steps, fin)
        return annealed

    if test_flag is False:
        lambda_parameter=linear_annealing(init=1, fin=100, step=global_train_step)
        x = (lambda_parameter*x).sigmoid()
    else:
        x = torch.where(x > 0, torch.ones_like(x), torch.zeros_like(x))
    return x



#######################################################################

class DummyPhiModule(nn.Module):
    def __init__(self):
        super(DummyPhiModule, self).__init__()

    def forward(self, x: Tensor) -> Tensor:
        return x.float()

    def test_time_forward(self, x):
        return self.forward(x).float()

    def regularization(self,x,global_train_step=0):
        return 0

# TODO: remove it?
class StraightPassingPhiModule(pl.LightningModule):
    def __init__(self,base_feature_extractor):
        super(StraightPassingPhiModule, self).__init__()
        self.base_feature_extractor = base_feature_extractor

        # the computation is just a identity function
        self.computation = torch.nn.Sequential()

    def forward(self, x, test_flag=False):
        x = self.base_feature_extractor(x,test_flag=test_flag)
        x = self.computation(x)
        if test_flag is False:
            output_representation = ImprovedSemanticHashing.apply(x)
        else:
            output_representation = ImprovedSemanticHashing.apply(x,True,self.device)
        return output_representation

    def test_time_forward(self, x):
        return self.forward(x,test_flag=True)

    def regularization(self,x,global_train_step=0):
        # rec_loss,regularization_loss_of_base_feature_extractor = self.base_feature_extractor.loss(x,global_train_step=global_train_step)

        rec_loss,regularization_loss_of_base_feature_extractor = self.base_feature_extractor.loss(x,global_train_step=global_train_step,test_flag=True)
        # here the test flag is on because we do not have to increase the increasing capacity

        # return 0,0
        return hypeperplane_regularization
        # return rec_loss,regularization_loss_of_base_feature_extractor


class HyperplanePhiModule(pl.LightningModule):
    def __init__(self,base_feature_extractor,latent_dim=10,n_concepts=8,binary_output=True,use_batch_norm=False,freeze_base=True):
        super(HyperplanePhiModule, self).__init__()
        self.base_feature_extractor = base_feature_extractor
        if freeze_base is True:
            self.base_feature_extractor.freeze()

        # self.computation = nn.Sequential(
        #     nn.Linear(latent_dim, 64),
        #     nn.ReLU(),
        #     nn.Linear(64, n_concepts),
        # )
        self.use_batch_norm=use_batch_norm
        if use_batch_norm==True:
            self.normalization=nn.BatchNorm1d(latent_dim)

        self.computation = nn.Linear(latent_dim,n_concepts,bias=True)

        self.binary_output=binary_output

        self.imbalance_ratio=None

    def forward(self, x, test_flag=False):
        x = self.base_feature_extractor(x,test_flag=test_flag).view(x.shape[0],-1)
        if self.use_batch_norm is True:
            self.normalized_x = self.normalization(x)
            x = self.normalized_x
        x = self.computation(x)

        self.x=x

        if self.binary_output is False:
            return x
        output_representation = ImprovedSemanticHashing(x,test_flag=test_flag,device=self.device)
        return output_representation

    def test_time_forward(self, x):
        return self.forward(x,test_flag=True)

    def regularization(self,x,global_train_step=0, concept_annotation=None):
        # regularization_loss_of_base_feature_extractor = self.base_feature_extractor.regularization()
        hypeperplane_regularization = self.get_concept_regularization(concept_annotation)
        return hypeperplane_regularization
        # return 0,0
        # return rec_loss,hypeperplane_regularization
        # return hypeperplane_regularization + regularization_loss_of_base_feature_extractor

    def get_concept_regularization(self,concept_annotation):
        if concept_annotation is None:
            return 0 * 0
        else:
            reg = self.compute_concept_loss(concept_annotation)
        return reg

    def compute_concept_loss(self,concept_annotation):
        learned_feature = self.x


        sigmoid_function = nn.Sigmoid()


        loss=0
        n_concepts=len(concept_annotation)

        if self.imbalance_ratio is None:
            self.imbalance_ratio = [1 for _ in range(n_concepts)]
            print("no imbalance_ratio!")

        loss_criterias = []
        for i in range(n_concepts):
            loss_criterias.append(nn.BCEWithLogitsLoss(pos_weight=torch.FloatTensor([self.imbalance_ratio[i]]).cuda()))

        for i in range(n_concepts):
            # loss_for_one_concept = loss_criteria(sigmoid_function(learned_feature[:,i]),concept_annotation[i].float().to(self.device))
            loss_for_one_concept = loss_criterias[i](learned_feature[:,i],concept_annotation[i].float())

            loss+= loss_for_one_concept

        # loss = loss / self.x.size()[0]
        loss = loss

        return loss

class GlobalPoolingHyperplanePhiModule(nn.Module):
    def __init__(self,base_feature_extractor,latent_dim=10,n_concepts=8,binary_output=True):
        super(HyperplanePhiModule, self).__init__()
        self.base_feature_extractor = base_feature_extractor

        # self.base_feature_extractor.freeze()

        self.computation = nn.Sequential(
            nn.BatchNorm1d(latent_dim),
            nn.Linear(latent_dim,n_concepts)
        )
        # self.glob
        self.binary_output=binary_output

    def forward(self, x, test_flag=False):
        x = self.base_feature_extractor(x,test_flag=test_flag).view(x.shape[0],-1)
        x = self.computation(x)

        if self.binary_output is False:
            return x
        if test_flag is False:
            output_representation = ImprovedSemanticHashing.apply(x)
        else:
            output_representation = ImprovedSemanticHashing.apply(x,True)
        return output_representation

    def test_time_forward(self, x):
        return self.forward(x,test_flag=True)
